#
#   Pyrex Scanner
#

#import pickle
import cPickle
pickle = cPickle

import os
import stat
import sys
from time import time

from Plex import *
from Plex.Errors import UnrecognizedInput
from Errors import CompileError, error

debug_scanner = 0
notify_lexicon_unpickling = 1

lexicon = None

reserved_words = [
  "global", "ctypedef", "cdef", "def", "class",
  "print", "del", "pass", "break", "continue", "return",
  "raise", "import", "exec", "try", "except", "finally",
  "while", "if", "elif", "else", "for", "in", "assert",
  "and", "or", "not", "is", "in", "lambda", "from"
]

#------------------------------------------------------------------

def modified_time(filename):
  s = os.stat(filename)
  return s[stat.ST_MTIME]

def lexicon_changed(source_file, lexicon_pickle):
  #print "source_file =", source_file ###
  #print "lexicon_pickle =", lexicon_pickle ###
  if not os.path.exists(lexicon_pickle):
    return 1
  source_time = modified_time(source_file)
  pickle_time = modified_time(lexicon_pickle)
  #print "source_time =", source_time ###
  #print "pickle_time =", pickle_time ###
  return source_time > pickle_time

def get_lexicon():
  global lexicon
  if not lexicon:
    dir = os.path.dirname(__file__)
    source_file = os.path.join(dir, "Scanning.py")
    lexicon_pickle = os.path.join(dir, "Lexicon.pickle")
    if not lexicon_changed(source_file, lexicon_pickle):
      if notify_lexicon_unpickling:
        t0 = time()
        print "Unpickling lexicon..."
      f = open(lexicon_pickle)
      lexicon = pickle.load(f)
      f.close()
      if notify_lexicon_unpickling:
        t1 = time()
        print "Done (%.2f seconds)" % (t1 - t0)
    else:
      t0 = time()
      print "Creating lexicon..."
      lexicon = make_lexicon()
      t1 = time()
      print "Done (%.2f seconds)" % (t1 - t0)
      t0 = time()
      print "Pickling lexicon..."
      f = open(lexicon_pickle, "w")
      pickle.dump(lexicon, f)
      f.close()
      t1 = time()
      print "Done (%.2f seconds)" % (t1 - t0)
  return lexicon
  
#------------------------------------------------------------------

class Method:

  def __init__(self, name):
    self.name = name
  
  def __call__(self, stream, text):
    return getattr(stream, self.name)(text)
    

def make_lexicon():

  letter = Any("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_")
  digit = Any("0123456789")
  octdigit = Any("01234567")
  hexdigit = Any("0123456789ABCDEFabcdef")
  #indentation = Bol + (Rep(Str(" ")) | Rep(Str("\t")))
  indentation = Bol + Rep(Any(" \t"))
  
  #resword = apply(Str, reserved_words)
  name = letter + Rep(letter | digit)
  decimal = Rep1(digit)
  intconst = decimal | (Str("0x") + Rep1(hexdigit))
  exponent = Any("Ee") + Opt(Any("+-")) + decimal
  fltconst = decimal + Opt(Str(".") + decimal) + Opt(exponent)
  imagconst = fltconst + Any("jJ")
  
  sq_string = (Str("'") + 
           Rep(AnyBut("\\\n'") | (Str("\\") + AnyChar)) + 
           Str("'"))
  dq_string = (Str('"') + 
           Rep(AnyBut('\\\n"') | (Str("\\") + AnyChar)) + 
           Str('"'))
  non_sq = AnyBut("'") | (Str('\\') + AnyChar)
  tsq_string = (
    Str("'''")
    + Rep(non_sq | (Str("'") + non_sq) | (Str("''") + non_sq)) 
    + Str("'''")
  )
  non_dq = AnyBut('"') | (Str('\\') + AnyChar)
  tdq_string = (Str('"""') +
           Rep(non_dq |
           (Str('"') + non_dq) |
           (Str('""') + non_dq)) + Str('"""'))
  stringlit = Opt(Any("rR")) + (sq_string | dq_string | tsq_string| tdq_string)

  beginstring = Opt(Any("rR")) + (Str("'") | Str('"') | Str("'''") | Str('"""'))
  two_oct = octdigit + octdigit
  three_oct = octdigit + octdigit + octdigit
  two_hex = hexdigit + hexdigit
  escapeseq = Str("\\") + (two_oct | three_oct | two_hex | AnyChar)

  bra = Any("([{")
  ket = Any(")]}")
  punct = Any(":,;+-*/|&<>=.%`~^")
  diphthong = Str("==", "<>", "!=", "<=", ">=", "<<", ">>", "**")
  spaces = Rep1(Any(" \t"))
  comment = Str("#") + Rep(AnyBut("\n"))
  escaped_newline = Str("\\\n")
  lineterm = Eol + Opt(Str("\n"))
  #lineterm = Str("\n")
  
  lexicon = Lexicon([
    #(resword, TEXT),
    (name, 'IDENT'),
    (intconst, 'INT'),
    (fltconst, 'FLOAT'),
    (imagconst, 'IMAG'),
    (punct | diphthong, TEXT),
    
    (bra, Method('open_bracket_action')),
    (ket, Method('close_bracket_action')),
    (lineterm, Method('newline_action')),

#		(stringlit, 'STRING'),
    (beginstring, Method('begin_string_action')),
    
    (comment, IGNORE),
    (spaces, IGNORE),
    (escaped_newline, IGNORE),
    
    State('INDENT', [
      (indentation + Opt(comment) + lineterm, IGNORE),
      (indentation, Method('indentation_action')),
      (Eof, Method('eof_action'))
    ]),
    
    State('SQ_STRING', [
      (escapeseq, 'ESCAPE'),
      (Rep1(AnyBut("'\"\n\\")), 'CHARS'),
      (Str('"'), 'CHARS'),
      (Eol + Str("\n"), Method('unclosed_string_action')),
      (Str("'"), Method('end_string_action')),
      (Eof, 'EOF')
    ]),
    
    State('DQ_STRING', [
      (escapeseq, 'ESCAPE'),
      (Rep1(AnyBut('"\n\\')), 'CHARS'),
      (Str("'"), 'CHARS'),
      (Eol + Str("\n"), Method('unclosed_string_action')),
      (Str('"'), Method('end_string_action')),
      (Eof, 'EOF')
    ]),
    
    State('TSQ_STRING', [
      (escapeseq, 'ESCAPE'),
      (Rep1(AnyBut("'\"\n\\")), 'CHARS'),
      (Any("'\""), 'CHARS'),
      (Eol + Str("\n"), 'NEWLINE'),
      (Str("'''"), Method('end_string_action')),
      (Eof, 'EOF')
    ]),
    
    State('TDQ_STRING', [
      (escapeseq, 'ESCAPE'),
      (Rep1(AnyBut('"\'\n\\')), 'CHARS'),
      (Any("'\""), 'CHARS'),
      (Eol + Str("\n"), 'NEWLINE'),
      (Str('"""'), Method('end_string_action')),
      (Eof, 'EOF')
    ]),
    
    (Eof, Method('eof_action'))
    ])

  return lexicon

#------------------------------------------------------------------

class PyrexScanner(Scanner):
  bracket_nesting_level = 0
  indentation_stack = None
  indentation_char = None

  def __init__(self, file, name):
    Scanner.__init__(self, get_lexicon(), file, name)
    self.build_resword_dict()
    self.indentation_stack = [0]
    self.begin('INDENT')
    self.next()
  
  def build_resword_dict(self):
    d = {}
    for word in reserved_words:
      d[word] = 1
    self.resword_dict = d
  
  def current_level(self):
    return self.indentation_stack[-1]

  def open_bracket_action(self, text):
    self.bracket_nesting_level = self.bracket_nesting_level + 1
    return text

  def close_bracket_action(self, text):
    self.bracket_nesting_level = self.bracket_nesting_level - 1
    return text

  def newline_action(self, text):
    if self.bracket_nesting_level == 0:
      self.begin('INDENT')
      self.produce('NEWLINE', '')
  
  string_states = {
    "'":   'SQ_STRING',
    '"':   'DQ_STRING',
    "'''": 'TSQ_STRING',
    '"""': 'TDQ_STRING'
  }
  
  def begin_string_action(self, text):
    if text[:1] in "rR":
      text = text[1:]
    self.begin(self.string_states[text])
    self.produce('BEGIN_STRING')
  
  def end_string_action(self, text):
    self.begin('')
    self.produce('END_STRING')
  
  def unclosed_string_action(self, text):
    self.end_string_action(text)
    self.error("Unclosed string literal")

  def indentation_action(self, text):
    self.begin('')
    # Indentation within brackets should be ignored.
    #if self.bracket_nesting_level > 0:
    #	return
    # Check that tabs and spaces are being used consistently.
    if text:
      c = text[0]
      #print "Scanner.indentation_action: indent with", repr(c) ###
      if self.indentation_char is None:
        self.indentation_char = c
        #print "Scanner.indentation_action: setting indent_char to", repr(c)
      else:
        if self.indentation_char <> c:
          self.error("Mixed use of tabs and spaces")
      if text.replace(c, "") <> "":
        self.error("Mixed use of tabs and spaces")
    # Figure out how many indents/dedents to do
    current_level = self.current_level()
    new_level = len(text)
    #print "Changing indent level from", current_level, "to", new_level ###
    if new_level == current_level:
      return
    elif new_level > current_level:
      #print "...pushing level", new_level ###
      self.indentation_stack.append(new_level)
      self.produce('INDENT', '')
    else:
      while new_level < self.current_level():
        #print "...popping level", self.indentation_stack[-1] ###
        self.indentation_stack.pop()
        self.produce('DEDENT', '')
      #print "...current level now", self.current_level() ###
      if new_level <> self.current_level():
        self.error("Inconsistent indentation")

  def eof_action(self, text):
    while len(self.indentation_stack) > 1:
      self.produce('DEDENT', '')
      self.indentation_stack.pop()
    self.produce('EOF', '')

  def next(self):
    try:
      #self.sy, self.systring = self.read()
      sy, systring = self.read()
    except UnrecognizedInput:
      self.error("Unrecognized character")
    if sy == 'IDENT' and systring in self.resword_dict:
      sy = systring
    self.sy = sy
    self.systring = systring
    if debug_scanner:
      _, line, col = self.position()
      if not self.systring or self.sy == self.systring:
        t = self.sy
      else:
        t = "%s %s" % (self.sy, self.systring)
      print "--- %3d %2d %s" % (line, col, t)
  
  def error(self, message, pos = None):
    if pos is None:
      pos = self.position()
    if self.sy == 'INDENT':
      error(pos, "Possible inconsistent indentation")
    raise error(pos, message)
    
  def expect(self, what, message = None):
    if self.sy == what:
      self.next()
    else:
      if message:
        self.error(message)
      else:
        self.error("Expected '%s'" % what)
    
  def expect_indent(self):
    self.expect('INDENT',
      "Expected an increase in indentation level")

  def expect_dedent(self):
    self.expect('DEDENT',
      "Expected a decrease in indentation level")

